﻿namespace HeuristicLab.BenchmarkSuite {
  using System;
  using System.Collections.Generic;
  using System.IO;
  using System.IO.Compression;
  using System.Linq;
  using System.Reflection;
  using System.Text.RegularExpressions;

  using HeuristicLab.BenchmarkSuite.ProblemData;

  using Microsoft.VisualBasic.FileIO;

  public abstract class BenchmarkSuiteDataDescriptor<TIn, TOut> : IBenchmarkSuiteDataDescriptor {
    //protected static IRandom rand = new FastRandom();

    private const string InstanceArchiveName = "HeuristicLab.BenchmarkSuite.Data.BenchmarkExamples.zip";
    private const string TestProblemPostfix = "Test.csv";
    private const string TrainingProblemPostfix = "Training.csv";
    private const string ResourcePath = @".*\.Data\.";
    private const string ExampleSeparator = ",";

    protected const char ArrayValueSeparator = ' ';
    protected static readonly char[] ArraySymbolTrim = { '[', ']' };

    public abstract string Name { get; }
    public abstract string Description { get; }

    protected abstract int InputCount { get; }
    protected abstract int OutputCount { get; }

    public abstract TIn ConvertInput(string[] input);
    public abstract TOut ConvertOutput(string[] input);

    public IData CreateData() {
      var training = this.GetTraining();
      var test = this.GetTest();

      return new Data<TIn, TOut>(training, test, 0, double.MaxValue);
    }

    public IPushData CreatePushData() {
      var training = this.GetTraining();
      var test = this.GetTest();

      return this.CreatePushData(training, test);
    }

    private Example<TIn, TOut>[] GetTest() {
      return this.ParseData(TestProblemPostfix).ToArray();
    }

    private Example<TIn, TOut>[] GetTraining() {
      return this.ParseData(TrainingProblemPostfix)/*.Shuffle(rand)*/.ToArray(); ;
    }

    public abstract IPushData CreatePushData(Example<TIn, TOut>[] training, Example<TIn, TOut>[] test);

    protected IEnumerable<Example<TIn, TOut>> ParseData(string postfix) {
      var type = this.GetType();
      var problemInstanceName = type.Name + postfix;

      using (var file = this.GetType().Assembly.GetManifestResourceStream(InstanceArchiveName))
      using (var archive = new ZipArchive(file, ZipArchiveMode.Read)) {
        var entry = archive.Entries.SingleOrDefault(x => x.Name == problemInstanceName);

        using (var parser = new TextFieldParser(entry.Open())) {
          parser.TextFieldType = FieldType.Delimited;
          parser.SetDelimiters(ExampleSeparator);

          while (!parser.EndOfData) {
            //Processing row
            var fields = parser.ReadFields();

            if (fields.Length != InputCount + OutputCount)
              throw new InvalidDataException("Number of values do not fit");

            yield return new Example<TIn, TOut> {
              Input = ConvertInput(fields.Take(this.InputCount).ToArray()),
              Output = ConvertOutput(fields.Skip(this.InputCount).ToArray())
            };
          }
        }
      }
    }

    protected static string GetResourceName(string fileName) {
      return Assembly
          .GetExecutingAssembly()
          .GetManifestResourceNames()
          .SingleOrDefault(x => Regex.Match(x, ResourcePath + fileName).Success);
    }

    protected static double[] ConvertDoubles(string input) {
      return ConvertMultiple(input, double.Parse);
    }

    protected static double ConvertDouble(string input) {
      return double.Parse(input);
    }

    protected static long[] ConvertIntegers(string input) {
      return ConvertMultiple(input, long.Parse);
    }

    protected static long ConvertInteger(string input) {

      return string.IsNullOrEmpty(input) ? default(long) : long.Parse(input);
    }

    protected static T[] ConvertMultiple<T>(string input, Func<string, T> converter) {
      return input
       .Trim(ArraySymbolTrim)
       .Split(ArrayValueSeparator)
       .Where(s => !string.IsNullOrWhiteSpace(s))
       .Select(converter)
       .ToArray();
    }

    protected static IEnumerable<string> SplitByNewLine(string input) {
      return input.Split(new[] { Environment.NewLine, "\n", "\r\n" }, StringSplitOptions.None);
    }

    private static string[] trueFormats = new[] { "true", "t", "1" };
    private static string[] falseFormats = new[] { "false", "f", "0" };

    protected static bool ConvertBool(string input) {
      var str = input.ToLower();

      if (trueFormats.Contains(str)) return true;
      if (falseFormats.Contains(str)) return false;
      throw new InvalidDataException(string.Format("Unable to parse {0} as boolean", str));
    }
  }
}
